/*
 * lock_server.c:
 *
 * Code for the external pager lock manger -- server side.
 * Erez Zadok.  Spring 1992. e-6118y.
 */

#include <stdio.h>
#include <mach.h>
#include <mach_error.h>
#include <mach/message.h>
#include <mach/mig_errors.h>
#include <servers/netname.h>
#include <assert.h>
#include <libc.h>

#include "xp_lock_manager.h"

#define NULL_QUEUE		(pending_queue_t *) NULL
#define NOMINAL_MSG_SIZE	32

struct pending_queue_t {
  msg_header_t request_elm, reply_elm;
  struct pending_queue_t *next;
};
typedef struct pending_queue_t pending_queue_t;

extern boolean_t locker_server(); /* from lockerServer.c */

static port_t server_port = PORT_NULL;

static msg_header_t last_reply, last_request;
static pending_queue_t *write_queue = NULL_QUEUE;
static pending_queue_t *head_reading_queue = NULL_QUEUE;
static pending_queue_t *head_pending_queue = NULL_QUEUE;

static int last_request_granted = FALSE;
static int current_state = STATE_NONE;
static int no_more_readers = FALSE;


/*
 * register_server with netname server.
 */
static void register_server()
{
  port_t port = PORT_NULL;

  while (port == PORT_NULL) {
    if (netname_look_up(name_server_port,"","NM_ACTIVE",&port) != KERN_SUCCESS)
      fprintf(stdout,"register_server: waiting for name server...\n");
    sleep(2);			/* doesn't work in threads */
  }

  port_deallocate(task_self(), port);

  if (netname_look_up(name_server_port, "", XP_LOCK_MANAGER_NAME,
		      &port) == KERN_SUCCESS) {
    fprintf(stdout,"register_server: %s is already registered\n",
	    XP_LOCK_MANAGER_NAME);
    exit(1);
  }

  if (port_allocate(task_self(), &server_port) != KERN_SUCCESS) {
    fprintf(stdout, "register_server: can't allocate server port\n");
    exit(1);
  }

  if (netname_check_in(name_server_port, XP_LOCK_MANAGER_NAME,
		       task_self(), server_port) != NETNAME_SUCCESS) {
    fprintf(stdout,"register_server: can't check in server name\n");
    exit(1);
  }

  printf("register_server: server name %s registered\n",
	 XP_LOCK_MANAGER_NAME);
}


static void unregister_server()
{
  if (netname_check_out(name_server_port, XP_LOCK_MANAGER_NAME,
			task_self()) != NETNAME_SUCCESS) {
    fprintf(stdout,"unregister_server: check out failed\n");
    exit(1);
  }
}


/*
 * Allocate a new queue element.
 */
static pending_queue_t *new_queue_elm()
{ pending_queue_t *tmp;

  puts("LOCK-PROTOCOL: new_queue_elm");
  tmp = (pending_queue_t *) calloc(1, NOMINAL_MSG_SIZE);
  assert(tmp != NULL_QUEUE);
  return tmp;
}

/*
 * Place a request in a queue of pending jobs.
 */
static void queue_request(last_request, last_reply)
     msg_header_t *last_request, *last_reply;
{ pending_queue_t *tmp;

  puts("LOCK-PROTOCOL: queue_request");
  tmp = new_queue_elm();
  bcopy((char *)last_request, (char *)&tmp->request_elm, NOMINAL_MSG_SIZE);
  bcopy((char *)last_reply, (char *)&tmp->reply_elm, NOMINAL_MSG_SIZE);
  tmp->next = head_pending_queue;
  head_pending_queue = tmp;
}

/*
 * Check if a requesting port is in the current reading queue.
 * Actually, since I allow multiple readers, this is not a queue but a
 * list.
 */
boolean_t in_reading_list(last_request)
     msg_header_t *last_request;
{ pending_queue_t *tmp;

  puts("LOCK-PROTOCOL: in_reading_list");
  tmp = head_reading_queue;
  while (tmp != NULL_QUEUE) {
    if (tmp->request_elm.msg_remote_port == last_request->msg_remote_port)
      return TRUE;
    tmp = tmp->next;
  }
  return FALSE;
}

/*
 * Add a reading client info into the currently reading queue.
 */
static void queue_reading(last_request, last_reply)
     msg_header_t *last_request, *last_reply;
{ pending_queue_t *tmp;

  puts("LOCK-PROTOCOL: queue_reading");
  tmp = new_queue_elm();
  printf("port number (5) = %d.\n", tmp->request_elm.msg_remote_port);
  bcopy((char *)last_request, (char *)&tmp->request_elm, NOMINAL_MSG_SIZE);
  printf("port number (6) = %d.\n", tmp->request_elm.msg_remote_port);
  bcopy((char *)last_reply, (char *)&tmp->reply_elm, NOMINAL_MSG_SIZE);
  printf("port number (7) = %d.\n", tmp->request_elm.msg_remote_port);
  tmp->next = head_reading_queue;
  head_reading_queue = tmp;
}

/*
 * Remove a requesting port from the current reading queue.
 */
static void dequeue_reading(last_request)
     msg_header_t *last_request;
{ pending_queue_t *tmp;

  puts("LOCK-PROTOCOL: dequeue_reading");
  tmp = head_reading_queue;

/*
 * case 1: first element in the queue is the one to dequeue.
 * works also if is only element in queue.
 */
  if (tmp->request_elm.msg_remote_port == last_request->msg_remote_port) {
    head_reading_queue = head_reading_queue->next;
    free(tmp);
    tmp = NULL_QUEUE;
    return;
  }

/*
 * case 2: there are two or more elements in the queue.  This means that
 * the element being queued is not the first one, so we have to know
 * which is the element that comes before it.
 */
  while (tmp->next != NULL_QUEUE) {
    if (tmp->next->request_elm.msg_remote_port ==
	last_request->msg_remote_port) {
      pending_queue_t *link;
      link = tmp->next;		/* store for later */
      tmp->next = link->next;	/* bypass link (null if last elm) */
      free(link);
      link = NULL_QUEUE;
      return;
    }
    tmp = tmp->next;
  }
}

/*
 * Check if the current reading queue is empty.
 */
static boolean_t is_empty_reading_queue()
{
  puts("LOCK-PROTOCOL: is_empty_reading_queue");
  return (head_reading_queue == NULL_QUEUE);
}

/*
 * Check if the this request is the last granted write access.
 */
static boolean_t only_one_writing(last_request)
     msg_header_t *last_request;
{
  puts("LOCK-PROTOCOL: only_one_writing");
  assert(write_queue != NULL_QUEUE);
  return (write_queue->request_elm.msg_remote_port ==
	  last_request->msg_remote_port);
}

/*
 * Empty the one and only entry in the write queue.
 */
static void clear_writing_queue()
{
  puts("LOCK-PROTOCOL: clear_writing_queue");
  assert(write_queue != NULL_QUEUE);
  free(write_queue);
  write_queue = NULL_QUEUE;
}

/*
 * Push only one entry on write queue.
 */
static void push_on_write_queue(last_request, last_reply)
     msg_header_t *last_request, *last_reply;
{
  puts("LOCK-PROTOCOL: push_on_write_queue");
  assert(write_queue == NULL_QUEUE);
  write_queue = new_queue_elm();
  bcopy((char *)last_request, (char *)&write_queue->request_elm,
	NOMINAL_MSG_SIZE);
  bcopy((char *)last_reply, (char *)&write_queue->reply_elm, NOMINAL_MSG_SIZE);
  write_queue->next = NULL_QUEUE;
}

/*
 * Check if the current pending queue is empty.
 */
static boolean_t is_empty_pending_queue()
{
  puts("LOCK-PROTOCOL: is_empty_pending_queue");
  return (head_pending_queue == NULL_QUEUE);
}

/*
 * Pop top of pending queue and fill into the out parameters.
 */
static void pop_pending_queue(request, reply_head)
     msg_header_t *request, *reply_head;
{ pending_queue_t *tmp;

  puts("LOCK-PROTOCOL: pop_pending_queue");
  assert(head_pending_queue != NULL_QUEUE);
  tmp = head_pending_queue;

/*
 * case 1: only one element in pending queue.
 */
  if (tmp->next == NULL_QUEUE) {
    bcopy((char *)&tmp->request_elm, (char *)request, NOMINAL_MSG_SIZE);
    bcopy((char *)&tmp->reply_elm, (char *)reply_head, NOMINAL_MSG_SIZE);
    free(tmp);
    tmp = head_pending_queue = NULL_QUEUE;
    return;
  }

/*
 * case 2: two or more elements in pending queue.
 * Find last element in linked list.
 */
  while (tmp->next->next != NULL_QUEUE)	/* find element second from end */
    tmp = tmp->next;
  bcopy((char *)&tmp->next->request_elm, (char *)request, NOMINAL_MSG_SIZE);
  bcopy((char *)&tmp->next->reply_elm, (char *)reply_head, NOMINAL_MSG_SIZE);
  free(tmp->next);
  tmp->next = NULL_QUEUE;  
}


kern_return_t ask_lock_manager(server_port, lck_ask, lck_get)
     port_t server_port;
     int lck_ask;
     int *lck_get;
{
  printf("server: server_port %d, lck_ask %d, lck_get %d.\n",
	 server_port, lck_ask, *lck_get);

  switch (lck_ask) {
  case XP_READ_LOCK:		/* asked for read lock */
    if (no_more_readers) {
      last_request_granted = FALSE;
      queue_request(&last_request, &last_reply);
      break;
    }
    if (current_state == STATE_READING  ||
	current_state == STATE_NONE) { /* ok for more to read */
      current_state = STATE_READING;
      last_request_granted = TRUE;
      no_more_readers = FALSE;	/* probably false already */
      queue_reading(&last_request, &last_reply);
    }
    if (current_state == STATE_WRITING) {
      no_more_readers = TRUE;	/* probably already true */
      last_request_granted = FALSE;
      queue_request(&last_request, &last_reply);
    }
    break;

  case XP_WRITE_LOCK:		/* asked for write lock */
    if (current_state == STATE_READING || /* don't grant request */
	current_state == STATE_WRITING) {
      no_more_readers = TRUE;
      last_request_granted = FALSE;
      queue_request(&last_request, &last_reply);
    }
    if (current_state == STATE_NONE) { /* grant request */
      current_state = STATE_WRITING;
      last_request_granted = TRUE;
      no_more_readers = TRUE;
      push_on_write_queue(&last_request, &last_reply);
    }
    break;

  case XP_READ_UNLOCK:
    if (current_state == STATE_NONE ||
	current_state == STATE_WRITING) { /* cannot be */
      printf("(1) MALICIOUS CLIENT!!!\n"); /* ignore malicious client! */
    }
    if (current_state == STATE_READING) {
      if (in_reading_list(&last_request)) {
	dequeue_reading(&last_request);
	last_request_granted = TRUE;
	if (is_empty_reading_queue()) /* then change state too */
	  current_state = STATE_NONE;
      } else {			/* another malicious hacker */
	printf("(2) MALICIOUS CLIENT!!!\n"); /* ignore malicious client! */
      }
    }
    break;

  case XP_WRITE_UNLOCK:
    if (current_state == STATE_NONE ||
	current_state == STATE_READING) { /* cannot be. only one writer */
      printf("(3) MALICIOUS CLIENT!!!\n"); /* ignore malicious client! */
    }
    if (current_state == STATE_WRITING) {
      if (only_one_writing(&last_request)) {
	clear_writing_queue();
	last_request_granted = TRUE;
	current_state = STATE_NONE;
	no_more_readers = FALSE;
      }
    }
    break;

  default:			/* this should not happen */
    printf("fatal error in decoding of external pager lock request\n");
    exit(1);
  }

  if (last_request_granted)	/* client sanity */
    *lck_get = lck_ask;

  return KERN_SUCCESS;
}


/*
 * Print the state of the lock manager.
 */
static void print_state()
{ pending_queue_t *tmp;

  printf("\tSTATE: ");
  switch (current_state) {
  case STATE_NONE:
    printf("none");
    break;
  case STATE_READING:
    printf("Reading");
    break;
  case STATE_WRITING:
    printf("Writing");
    break;
  default:
    printf("UNKNOWN");
    break;
  }
  printf("\n\tLAST_REQUEST_GRANTED: %s",
	 last_request_granted ? "TRUE" : "FALSE");
  printf("\n\tNO_MORE_READERS: %s",
	 no_more_readers ? "TRUE" : "FALSE");
  printf("\n\tPENDING_QUEUE: ");
  tmp = head_pending_queue;
  while (tmp != NULL_QUEUE) {
    printf("%d ", tmp->request_elm.msg_remote_port);
    printf("\tmessage size was %d and struct is %d\n",
	   tmp->request_elm.msg_size, sizeof(msg_header_t));
    tmp = tmp->next;
  }
  printf("\n\tREADING_QUEUE: ");
  tmp = head_reading_queue;
  while (tmp != NULL_QUEUE) {
    printf("%d ", tmp->request_elm.msg_remote_port);
    printf("\tmessage size was %d and struct is %d\n",
	   tmp->request_elm.msg_size, sizeof(msg_header_t));
    tmp = tmp->next;
  }
  printf("\n\tWRITE_QUEUE: ");
  tmp = write_queue;
  while (tmp != NULL_QUEUE) {
    printf("%d ", tmp->request_elm.msg_remote_port);
    printf("\tmessage size was %d and struct is %d\n",
	   tmp->request_elm.msg_size, sizeof(msg_header_t));
    tmp = tmp->next;
  }
  printf("\n\n");
}


int main()
{
  int requestbuf[NOMINAL_MSG_SIZE/sizeof(int)];
  int replybuf[NOMINAL_MSG_SIZE/sizeof(int)];
  msg_header_t *request = (msg_header_t *) requestbuf;
  death_pill_t *reply = (death_pill_t *) replybuf;
  msg_return_t mr;

  register_server();

  write_queue = NULL_QUEUE;
  head_reading_queue = NULL_QUEUE;
  head_pending_queue = NULL_QUEUE;

  fprintf(stdout,"server_loop: starting\n");
  while(1) {
    print_state();
    assert(request != (msg_header_t *) NULL);
    request->msg_size = sizeof requestbuf;
    request->msg_local_port = server_port;

    if (last_request_granted && !is_empty_pending_queue()) {
      printf("POPPING Q (msg_rec) if not empty: two print_states()s:\n");
      print_state();
      pop_pending_queue(request, &reply->Head);
      print_state();
      goto process_again;	/* eat me! */
    }

    printf("server_loop: about to receive msg\n");
    mr = msg_receive(request, RCV_INTERRUPT, 0);
    if (mr != RCV_SUCCESS) {
      mach_error("lock_server", mr);
      port_deallocate(task_self(), server_port);
      exit(1);
    }
    fprintf(stdout,"server_loop: after receive\n");

    printf("server_loop: before ports %d %d %d %d\n",
	   request->msg_local_port, request->msg_remote_port,
	   reply->Head.msg_local_port, reply->Head.msg_remote_port);

    if (request->msg_local_port == task_notify()) {
      printf("ignoring notification from kernel\n");
      continue;
    }

  process_again:

    printf("bcopying last reply and request\n");
    printf("port number (1) = %d.\n", last_request.msg_remote_port);
    bcopy((char *)&reply->Head, (char *)&last_reply, NOMINAL_MSG_SIZE);
    bcopy((char *)request, (char *)&last_request, NOMINAL_MSG_SIZE);
    printf("port number (2) = %d.\n", last_request.msg_remote_port);
    printf("calling locker_server: request=%x, reply->head=%x\n",
	   request, reply->Head);
    locker_server(request, &reply->Head);
    printf("port number (3) = %d.\n", last_request.msg_remote_port);
    printf("back from locker_server: request=%x, reply->Head=%x\n",
	   request, reply->Head);
    if (last_request_granted) { /* was last request granted? */
				/* send reply back */
      if ((reply->Head.msg_remote_port != PORT_NULL) &&
	  (reply->RetCode != MIG_NO_REPLY)) {
				/* don't block if port is full */
	msg_send(&reply->Head, SEND_TIMEOUT | SEND_INTERRUPT, 0);
      }
      if (!is_empty_pending_queue()) {
	printf("POPPING QUEUE if not empty: two print_states()s:\n");
	print_state();
	pop_pending_queue(request, &reply->Head);
	print_state();
	goto process_again;	/* eat me! */
      }
    }
    printf("port number (4) = %d.\n", last_request.msg_remote_port);
    fprintf(stdout,"server_loop: at end of loop\n");
  }

  unregister_server();
}
